#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

//Mode0
//BG0 
//BG2 g
//BG3 pl

//---------------------------------------------------------------------------
#define PANEL_LENGTH_X			9	//pl̃TCY
#define PANEL_LENGTH_Y			9
#define PANEL_TYPE_READ			1	//pl̃^Cv  }bvf[^ɊɊi[Ă鐔l
#define PANEL_TYPE_INPUT		2

//Q[̐is
#define STATE_INIT				1
#define STATE_TITLE				2
#define STATE_STAGE_DISP		3
#define STATE_STAGE_INPUT		4
#define STATE_STAGE_PROC		5
#define STATE_END_PROC			6
#define STATE_CLEAR_PROC		7

//BG̃^C
#define BG_TILE_BLACK			0
#define BG_TILE_NUMBER1			2
#define BG_TILE_NUMBER2			4
#define BG_TILE_NUMBER3			6
#define BG_TILE_NUMBER4			8
#define BG_TILE_NUMBER5			10
#define BG_TILE_NUMBER6			12
#define BG_TILE_NUMBER7			14
#define BG_TILE_NUMBER8			16
#define BG_TILE_NUMBER9			18
#define BG_TILE_EMPTY			20

//̓L[̃EFCg
#define KEY_WAIT				3

//---------------------------------------------------------------------------

//pl
typedef struct {
	u8  type;
	u8  value;
	u8  answer;
	u16 expect;
	u8  expectCnt;
} ST_PANEL;

//BG
typedef struct {
	u8 Chr;
	u8 Map;
	u8 Col;
} ST_BG;

//J[\
typedef struct {
	s16 x;
	s16 y;
	u16 cx;
	u16 cy;
	s16 moveX;
	s16 moveY;
	s16 minX;		//ړ
	s16 minY;
	s16 maxX;
	s16 maxY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
	u8  key;		//ړp
	u16 keyCnt;		//L[s[hp
	u16 keyOld;		//L[s[hp
} ST_CURSOR;

//---------------------------------------------------------------------------
ST_BG     BG0;
ST_BG     BG2;
ST_BG     BG3;
ST_CURSOR Cursor;
ST_PANEL  Stage[PANEL_LENGTH_X][PANEL_LENGTH_Y];

u8 State;			//Q[̐istO
u8 StageCnt;

//---------------------------------------------------------------------------
void StateInit();					//
void StateTitle();					//^Cg
void StateStageDisp();				//Ֆʂ̕\
void StateStageInput();				//pl
void StateStageProc();				//plI
void StateEndProc();				//I
void StateClearProc();				//NA

void InitCursor();
void CursorClear(ST_CURSOR* p);

void StageInit();
void StageClear();
void StageLoad();
void StageSolve();
bool NextStage();
bool IsStageLoadFindBeginMessage();
bool IsStageEnd();

void DrawBG(u8 x, u8 y, u8 tile, u8 map, u8 col);
void DrawBGPanel(u8 x, u8 y);

u16  GetExpectTate(u8);
u16  GetExpectYoko(u8);
u16  GetExpectArea(u8,u8);
bool SetSolveNum(u8,u8);

void Input(ST_CURSOR* p);
void MoveCursor(ST_CURSOR* p);
u8   ChkMoveCursor(ST_CURSOR* p);
void DrawSprCursor(ST_CURSOR* p);
bool RepeatKey(u8);					//L[̉ςȂ
//---------------------------------------------------------------------------
int main(void)
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:         StateInit();        break;
			case STATE_TITLE:        StateTitle();       break;
			case STATE_STAGE_DISP:   StateStageDisp();   break;
			case STATE_STAGE_INPUT:  StateStageInput();  break;
			case STATE_STAGE_PROC:   StateStageProc();   break;
			case STATE_END_PROC:     StateEndProc();     break;
			case STATE_CLEAR_PROC:   StateClearProc();   break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	//EFCgRg[
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;

	//L[
	KeyInit();
	Cursor.keyCnt = 0;
	Cursor.keyOld = 0;

	//BG
	//BG0 
	//BG2 g
	//BG3 pl

	BG0.Chr = 1;
	BG0.Map = 0;
	BG0.Col = 0;

	BG2.Chr = 2;	//BG3Ɠ
	BG2.Map = 2;
	BG2.Col = 1;	//BG3Ɠ

	BG3.Chr = 2;
	BG3.Map = 1;
	BG3.Col = 1;

	REG_BG0CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG0.Chr) | BG_MAPBASE(BG0.Map) | 0);

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.Chr) | BG_MAPBASE(BG2.Map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.Chr) | BG_MAPBASE(BG3.Map) | 0);

	BGSetPal16  (BG0.Col, (u16*)&BGFont_palette);
	BGSetChrData(BG0.Chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);

	BGSetPal16  (BG3.Col, (u16*)&bg_pal_bin);
	BGSetChrData(BG3.Chr, (u16*)&bg_img_bin, 4096);

	BGSetMapData(BG2.Map, (u16*)&bg_waku_img_bin, 2048);
	BGSetPalMap (BG2.Map, 2048, BG2.Col);

	//Sprite
	SpriteInit();
	SpriteSetPal16((u16*)&spr_pal_bin);
	SpriteSetData ((u16*)&spr_img_bin, 2048);

	//Stage
	StageInit();
	NextStage();	//1ʂǂݍ݂܂

	//Cursor
	InitCursor();

	SetMode(MODE_0 | BG0_ENABLE | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);

	//is^Cg\
	State = STATE_TITLE;
}
//-----------------------------------------------------------------------------
void StateTitle()
{
	//ubNC
	//vbVX^[gi_Łj{^܂ŃEFCg
	//ubNAEg
	
	State = STATE_STAGE_DISP;
}
//---------------------------------------------------------------------------
void StateStageDisp()
{
	u8 x,y;

	for(y=0; y<PANEL_LENGTH_Y; y++)
	{
		for(x=0; x<PANEL_LENGTH_X; x++)
		{
			DrawBGPanel(x, y);
		}
	}
	//isL[͂
	State = STATE_STAGE_INPUT;
}
//---------------------------------------------------------------------------
void StateStageInput()
{
	Input(&Cursor);
}
//---------------------------------------------------------------------------
void DrawBGPanel(u8 x, u8 y)
{
	u8 value = Stage[x][y].value;
	u8 type  = Stage[x][y].type;
	u8 bgX   = x*2;		//WBGW֕ϊ
	u8 bgY   = y*2;
	u8 tile  = 0x00;

	switch(value)
	{
		case 0: tile = BG_TILE_EMPTY;   break;
		case 1: tile = BG_TILE_NUMBER1; break;
		case 2: tile = BG_TILE_NUMBER2; break;
		case 3: tile = BG_TILE_NUMBER3; break;
		case 4: tile = BG_TILE_NUMBER4; break;
		case 5: tile = BG_TILE_NUMBER5; break;
		case 6: tile = BG_TILE_NUMBER6; break;
		case 7: tile = BG_TILE_NUMBER7; break;
		case 8: tile = BG_TILE_NUMBER8; break;
		case 9: tile = BG_TILE_NUMBER9; break;
		default:
			DebugMessage((u8*)"Err: DrawBGPanel value");
			break;
	}

	//ƋꂵEEE
	if(type == PANEL_TYPE_INPUT && tile != BG_TILE_EMPTY)
	{
		tile += 64;
	}
	
	DrawBG(bgX, bgY, tile, BG3.Map, BG3.Col);
}
//---------------------------------------------------------------------------
void DrawBG(u8 x, u8 y, u8 tile, u8 map, u8 col)
{
	BGDrawChar(map, x+0, y+0, col, tile+0 );
	BGDrawChar(map, x+1, y+0, col, tile+1 );
	BGDrawChar(map, x+0, y+1, col, tile+32);
	BGDrawChar(map, x+1, y+1, col, tile+33);
}
//---------------------------------------------------------------------------
//L[s[h
bool RepeatKey(u8 Key)
{
	if(Key == 0x00 || Cursor.keyOld != Key)
	{
		Cursor.keyOld = Key;
		Cursor.keyCnt = 0;
		return TRUE;
	}

	if(Cursor.keyCnt <= KEY_WAIT)
	{
		Cursor.keyCnt++;
		return TRUE;
	}

	Cursor.keyCnt = 0;
	return FALSE;
}
//---------------------------------------------------------------------------
void StageInit()
{
	StageCnt = 0;
	SeekSetChr((u8*)&sudokuvol2_data_bin, 0x6957, 'b');
}
//---------------------------------------------------------------------------
bool NextStage()
{
	//̃Xe[W񂪂邩mF܂
	if( IsStageLoadFindBeginMessage() == FALSE )
	{
		return FALSE;
	}

	StageClear();
	StageLoad();
	StageSolve();

	return TRUE;
}
//---------------------------------------------------------------------------
//Xe[WϐɊi[܂
void StageLoad()
{
	StageCnt++;

	u8 x, y, num;
	u8* pData = SeekGetNow() + 6;	//6͕␳ł

	for(y=0; y<PANEL_LENGTH_Y; y++)
	{
		for(x=0; x<PANEL_LENGTH_X; x++)
		{
			num = *(pData+1) == 0x2d ? 0 : *(pData+1) - 0x30;

			Stage[x][y].value  = num;
			Stage[x][y].answer = num;
			Stage[x][y].type   = num == 0 ? PANEL_TYPE_INPUT : PANEL_TYPE_READ;
			pData += 3;
		}
	}
}
//---------------------------------------------------------------------------
//f[^̓oBbZ[Wubegin.v܂ŃItZbg̈ړ܂
bool IsStageLoadFindBeginMessage()
{
	while(SeekHasMore() == TRUE)
	{
		u8* pData = SeekNext();

		if( *(pData+0) != 'b' ) continue;
		if( *(pData+1) != 'e' ) continue;
		if( *(pData+2) != 'g' ) continue;
		if( *(pData+3) != 'i' ) continue;
		if( *(pData+4) != 'n' ) continue;

		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
//ړ͈ړ\`FbN܂
u8 ChkMoveCursor(ST_CURSOR* p)
{
	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->key & KEY_UP   ) ty1 -= p->moveY;
	if(p->key & KEY_DOWN ) ty1 += p->moveY;
	if(p->key & KEY_LEFT ) tx1 -= p->moveX;
	if(p->key & KEY_RIGHT) tx1 += p->moveX;

	if(tx1< p->minX || ty1< p->minY) return FALSE;
	if(tx1>=p->maxX || ty1>=p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
//J[\̈ړ
void MoveCursor(ST_CURSOR* p)
{
	if(p->key & KEY_UP   ) p->y-=p->moveY;
	if(p->key & KEY_DOWN ) p->y+=p->moveY;
	if(p->key & KEY_LEFT ) p->x-=p->moveX;
	if(p->key & KEY_RIGHT) p->x+=p->moveX;
}
//---------------------------------------------------------------------------
void Input(ST_CURSOR* p)
{
	u8 Key = KeyInput();

	//s[h
	if( RepeatKey(Key) == TRUE )
	{
		return;
	}

	p->key = 0x00;

	if(Key & KEY_UP   ) p->key |= KEY_UP;
	if(Key & KEY_DOWN ) p->key |= KEY_DOWN;
	if(Key & KEY_LEFT ) p->key |= KEY_LEFT;
	if(Key & KEY_RIGHT) p->key |= KEY_RIGHT;

	if( ChkMoveCursor(p) == TRUE )
	{
		MoveCursor(p);
		DrawSprCursor(p);
	}

	if(Key & KEY_A)
	{
		State = STATE_STAGE_PROC;
	}
}
//---------------------------------------------------------------------------
void StageClear()
{
	u8 x, y;

	for(y=0; y<PANEL_LENGTH_Y; y++)
	{
		for(x=0; x<PANEL_LENGTH_X; x++)
		{
			Stage[x][y].value     = 0;
			Stage[x][y].answer    = 0;
			Stage[x][y].type      = PANEL_TYPE_INPUT;
			Stage[x][y].expect    = 0x0000;
			Stage[x][y].expectCnt = 0;
		}
	}
}
//---------------------------------------------------------------------------
void DrawSprCursor(ST_CURSOR* p)
{
	SpriteMove(p->sprNo, p->x, p->y);
}
//---------------------------------------------------------------------------
void InitCursor()
{
	Cursor.x       = 0;
	Cursor.y       = 0;
	Cursor.cx      = 16;
	Cursor.cy      = 16;
	Cursor.moveX   = 16;
	Cursor.moveY   = 16;
	Cursor.minX    = 0;
	Cursor.minY    = 0;
	Cursor.maxX    = 16*9;
	Cursor.maxY    = 16*9;
	Cursor.sprNo   = 0;
	Cursor.sprTile = 0;
	Cursor.sprSize = SP_SIZE_16;
	Cursor.sprForm = SP_SQUARE;
	Cursor.key     = 0x0000;
	Cursor.keyCnt  = 0x00;
	Cursor.keyOld  = 0x00;

	SpriteSetSize(Cursor.sprNo, Cursor.sprSize, Cursor.sprForm, SP_COLOR_16);
	SpriteSetChr (Cursor.sprNo, Cursor.sprTile);
	SpriteMove   (Cursor.sprNo, Cursor.x, Cursor.y);
}
//---------------------------------------------------------------------------
void StageSolve()
{
	u8   x,y,type;
	u8   cnt      = 0;
	bool backflag = FALSE;

	while(cnt < PANEL_LENGTH_X * PANEL_LENGTH_Y)
	{
		x    = cnt % PANEL_LENGTH_X;
		y    = cnt / PANEL_LENGTH_Y;
		type = Stage[x][y].type;

		//ǂݍݐl̏ꍇ
		if(type == PANEL_TYPE_READ)
		{
			if(backflag) cnt--;
			else         cnt++;
			continue;
		}

		//li[܂
		if( SetSolveNum(x, y) == TRUE )
		{
			//A̍Wֈړ
			cnt++;
			backflag = FALSE;
		}
		else
		{
			//sAO̍Wֈړ
			cnt--;
			backflag = TRUE;
		}
	}
}
//-----------------------------------------------------------------------------
bool SetSolveNum(u8 x, u8 y)
{
	ST_PANEL* p = &Stage[x][y];

	//Wɒu鐔lu16̃rbgPʂŋ߂܂uxxxx xx98 7654 321xv
	p->expect  = 0x0000;
	p->expect |= GetExpectTate(x);
	p->expect |= GetExpectYoko(y);
	p->expect |= GetExpectArea(x, y);

	while(p->expectCnt++ < 9)
	{
		if( !(p->expect & (0x0001 << p->expectCnt)) )
		{
			p->answer = p->expectCnt;
			return TRUE;
		}
	}

	p->answer    = 0;
	p->expectCnt = 0;
	return FALSE;
}
//-----------------------------------------------------------------------------
//̗Ɋi[Ă鐔l߂lɂ܂
u16 GetExpectTate(u8 x)
{
	u8  y;
	u16 retFlag = 0x0000;

	for(y=0; y<PANEL_LENGTH_Y; y++)
	{
		if(Stage[x][y].answer != 0)
		{
			retFlag |= 0x0001 << Stage[x][y].answer;
		}
	}
	return retFlag;
}
//-----------------------------------------------------------------------------
//c̗Ɋi[Ă鐔l߂lɂ܂
u16 GetExpectYoko(u8 y)
{
	u8  x;
	u16 retFlag = 0x0000;
	
	for(x=0; x<PANEL_LENGTH_X; x++)
	{
		if(Stage[x][y].answer != 0)
		{
			retFlag |= 0x0001 << Stage[x][y].answer;
		}
	}
	return retFlag;
}
//-----------------------------------------------------------------------------
u16 GetExpectArea(u8 x, u8 y)
{
	u8 i,j,x0,y0;
	u16 retFlag = 0x0000;

	x0 = (x / 3) * 3;
	y0 = (y / 3) * 3;

	for(i=0; i<3; i++)
	{
		for(j=0; j<3; j++)
		{
			if(Stage[x0+i][y0+j].answer != 0)
			{
				retFlag |= 0x0001 << Stage[x0+i][y0+j].answer;
			}
		}
	}
	return retFlag;
}
//---------------------------------------------------------------------------
bool IsStageEnd()
{
	u8 x, y;

	for(y=0; y<PANEL_LENGTH_Y; y++)
	{
		for(x=0; x<PANEL_LENGTH_X; x++)
		{
			if(Stage[x][y].value != Stage[x][y].answer)
			{
				return FALSE;
			}
		}
	}
	return TRUE;
}
//---------------------------------------------------------------------------
void StateStageProc()
{
	u8 x        = Cursor.x / Cursor.moveX;
	u8 y        = Cursor.y / Cursor.moveY;
	ST_PANEL* p = &Stage[x][y];

	if(p->type == PANEL_TYPE_READ)
	{
		State = STATE_STAGE_INPUT;
		return;
	}

	p->value = ++p->value % 10;
	DrawBGPanel(x, y);

	if( IsStageEnd() == TRUE )
	{
		BGDrawStr(BG0.Map, 20, 2, (u8*)"Clear!!");
		State = STATE_CLEAR_PROC;
	}
	else
	{
		State = STATE_STAGE_INPUT;
	}
}
//---------------------------------------------------------------------------
void StateEndProc()
{
	//͂܂
}
//---------------------------------------------------------------------------
void StateClearProc()
{
	//STARTL[܂ŉȂ
	u8 Key = KeyInput();
	if(!(Key & KEY_START))
	{
		return;
	}

	if( NextStage() == TRUE )
	{
		CursorClear(&Cursor);
		BGDrawCls(BG0.Map, 0x00);
		State = STATE_STAGE_DISP;
	}else
	{
		BGDrawStr(BG0.Map, 18, 3, (u8*)"ALL Clear!!");
		State = STATE_END_PROC;
	}
}
//---------------------------------------------------------------------------
void CursorClear(ST_CURSOR* p)
{
	p->x = 0;
	p->y = 0;
	DrawSprCursor(p);
}
